from typing import Optional, List, Literal

from pydantic import Field, field_validator, ValidationInfo
from anssformats.formatbasemodel import CustomDT, FormatBaseModel
from anssformats.geojson import PointGeometry


class HypocenterProperties(FormatBaseModel):
    """A class holding the hypocenter specific custom properties for a geojson point feature

    originTime: required datetime containing the origin time of the hypocenter

    latitudeError: optional float containing the error of the latitude of this
        hypocenter in kilometers

    longitudeError: optional float containing the error of the longitude of this
        hypocenter in kilometers

    depthError: optional float containing the error of the depth of this hypocenter in
        kilometers

    timeError: optional float containing the error of the origin time of this hypocenter
        in seconds
    """

    originTime: CustomDT
    latitudeError: Optional[float] = None
    longitudeError: Optional[float] = None
    depthError: Optional[float] = None
    timeError: Optional[float] = None


class Hypocenter(FormatBaseModel):
    """A conversion class used to create, parse, and validate geojson Hypocenter data as part of
    detection data.

    type: string containing the type of this geojson

    geometry: PointGeometry object containing the geojson geometry for this feature

    properties: HypocenterProperties object containing the hypocenter properties
    """

    type: str = "Feature"

    geometry: PointGeometry
    properties: HypocenterProperties



from typing import Optional

from pydantic import Field

from anssformats.formatbasemodel import FormatBaseModel
from anssformats.source import Source


class Magnitude(FormatBaseModel):
    """A conversion class used to create, parse, and validate Magnitude data as part
    of detection data.

    Attributes
    ----------

    value: float containing the magnitude value

    type: string containing the magnitude type

    error: optional float containing the associated magnitude error (if any)

    probability: optional float containing the associated magnitude probability (if any)

    id: optional string containing a unique identifier for this magnitude

    source: optional Source object containing the source of the magnitude
    """

    value: float = Field(None, ge=-2.0, le=10.0)
    type: str

    error: Optional[float] = Field(None, ge=0.0)
    probability: Optional[float] = Field(None, ge=0.0, le=100.0)
    id: Optional[str] = None
    source: Optional[Source] = None



from typing import Optional, Union, List

from pydantic import Field

from anssformats.formatbasemodel import FormatBaseModel
from anssformats.source import Source as SourceFormat


class Prediction(FormatBaseModel):
    """A generic prediction from an analytical model (AI/ML, statistical, rule-based, etc.)

    This class provides a flexible structure for representing any type of prediction
    with associated metrics and model provenance.

    Attributes
    ----------
    label: string identifying what is being predicted (e.g., "phase", "magnitude",
        "eventType", "custom_attribute")

    value: the predicted value - can be string, number, or structured object

    probability: optional float [0.0-1.0] containing the probability of this prediction

    metrics: optional dict containing additional prediction metrics such as confidence,
        uncertainty quantification (std, ranges, credible intervals), entropy, etc.
        Structure is flexible to support any model-specific metrics.

    modelID: optional string identifying which model made this prediction

    modelVersion: optional string containing the version of the model

    source: optional Source object containing the source/author of the
    """

    label: str = Field(
        ...,
        description="What is being predicted (e.g., phase, magnitude, eventType, distance)",
    )
    value: Union[str, float, int, list, dict] = Field(
        ..., description="The predicted value"
    )
    probability: Optional[float] = Field(
        None, ge=0.0, le=1.0, description="Probability of this prediction [0.0-1.0]"
    )
    metrics: Optional[dict] = Field(
        None,
        description="Additional prediction metrics (confidence, uncertainty, entropy, etc.)",
    )
    modelID: Optional[str] = Field(
        None, description="Identifier for the model that made this prediction"
    )
    modelVersion: Optional[str] = Field(None, description="Version of the model")
    source: Optional[SourceFormat] = Field(
        None, description="Source/author of the model"
    )


class Analytics(FormatBaseModel):
    """A conversion class used to create, parse, and validate analytical information
    from models such as pickers, analytical models, AI/ML, etc.

    This class provides an extensible structure supporting multiple models, arbitrary
    prediction labels, and custom extensions.

    Attributes
    ----------

    predictions: optional list of Prediction objects containing predictions from one
        or more models. Each prediction can represent any type of analytical output
        with associated confidence metrics and model provenance.

    extensions: optional dict containing custom key-value pairs for experimental code and debugging.

    """

    predictions: Optional[List[Prediction]] = Field(
        None, description="Array of predictions from one or more analytical models"
    )

    extensions: Optional[dict] = Field(
        None, description="Custom key-value pairs for experimental or debuging data"
    )



from typing import Optional

from anssformats.formatbasemodel import FormatBaseModel


class Filter(FormatBaseModel):
    """A conversion class used to create, parse, and validate filter data as part of
    detection data.

    Attributes
    ----------

    type: optional string containing the type of filter

    highPass: optional float containing the high pass frequency

    lowPass: optional float containing the low pass frequency

    units: optional string containing the filter frequency units
    """

    type: Optional[str] = None
    highPass: Optional[float] = None
    lowPass: Optional[float] = None
    units: Optional[str] = None



from typing import List, Literal, Optional

from pydantic import Field

from anssformats.amplitude import Amplitude
from anssformats.association import Association
from anssformats.analytics import Analytics
from anssformats.filter import Filter
from anssformats.formatbasemodel import CustomDT, FormatBaseModel
from anssformats.channel import Channel, ChannelProperties
from anssformats.source import Source


class Pick(FormatBaseModel):
    """A conversion class used to create, parse, and validate pick detection data.

    type: string identifying this message as a pick

    id: string containing a unique identifier for this pick

    channel: Channel object containing the station where the pick was made

    source: Source object containing the source of the pick

    time: datetime containing the arrival time of the phase that was picked

    phase: optional string containing the seismic phase that was picked

    polarity: optional string containing the phase polarity; allowed strings are "up"
        and "down"

    onset: optional string containing the phase onset; allowed strings are "impulsive",
        "emergent", and "questionable"

    pickerType: optional string containing the type of picker; allowed strings are "manual",
        "raypicker", "filterpicker", "earthworm", and "other"

    filterInfo: optional list of Filter objects containing the filter frequencies when the
        pick was made

    amplitude: optional Amplitude object containing the amplitude associated with the pick

    associationInfo: optional Association object containing the association information
        if this pick is used as data in a Detection

    AnalyticsInfo: optional analytics object containing model output attached to this pick
    """

    type: Literal["Pick"]
    id: str

    channel: Channel
    source: Source
    time: CustomDT

    phase: Optional[str] = Field(None, pattern=r"^[A-Za-z]+$")
    polarity: Optional[Literal["up", "down"]] = None
    onset: Optional[Literal["impulsive", "emergent", "questionable"]] = None
    pickerType: Optional[
        Literal["manual", "raypicker", "filterpicker", "earthworm", "other"]
    ] = None

    filterInfo: Optional[List[Filter]] = None
    amplitudeInfo: Optional[Amplitude] = None

    associationInfo: Optional[Association] = None
    analyticsInfo: Optional[Analytics] = None



from typing import Optional, List, Literal

from pydantic import Field, field_validator, ValidationInfo
from anssformats.formatbasemodel import FormatBaseModel


class PointGeometry(FormatBaseModel):
    """A class holding a geojson point geometry

    type: string containing the type of this geometry

    coordinates: List of floats containing the longitude in degrees, latitude in degrees, and elevation in meters or depth in kilometers, in that order
    """

    type: str = "Point"
    coordinates: List[float]

    # check that coordinates are valid
    @field_validator("coordinates")
    @classmethod
    def validate_coordinates(
        cls, value: List[float], info: ValidationInfo
    ) -> List[float]:
        if value is None:
            raise ValueError("Missing coordinates")

        if len(value) != 3:
            raise ValueError("Incomplete coordinates")

        # longitude
        if value[0] < -180.0 or value[0] > 180.0:
            raise ValueError("Longitude coordinate out of valid range")

        # latitude
        if value[1] < -90.0 or value[1] > 90.0:
            raise ValueError("Latitude coordinate out of valid range")

        # don't validate elevation/depth
        # value[2]

        return value



from typing import Literal, Optional

from anssformats.formatbasemodel import FormatBaseModel


class EventType(FormatBaseModel):
    """A conversion class used to create, parse, and validate source data as part of
    detection data. Detection formats supports a subset of the QuakeML 1.2 event types
    that are automatically classifiable.

    Attributes
    ----------

    type: optional string containing the type of detection that was found; allowed type
        strings are: "Earthquake", "MineCollapse", "NuclearExplosion", "QuarryBlast",
        "InducedOrTriggered", "RockBurst", "FluidInjection", "IceQuake", and
        "VolcanicEruption"

    certainty: optional string containing the certainty of the event type; allowed
        strings are "Suspected" and "Confirmed"
    """

    type: Optional[
        Literal[
            "Earthquake",
            "MineCollapse",
            "NuclearExplosion",
            "QuarryBlast",
            "InducedOrTriggered",
            "RockBurst",
            "FluidInjection",
            "IceQuake",
            "VolcanicEruption",
        ]
    ] = None

    certainty: Optional[Literal["Suspected", "Confirmed"]] = None



from typing import Optional, List, Literal

from pydantic import Field, field_validator, ValidationInfo
from anssformats.formatbasemodel import FormatBaseModel
from anssformats.geojson import PointGeometry


class ChannelProperties(FormatBaseModel):
    """A class holding the channel specific custom properties for a geojson point feature

    Station: string containing the station code

    Channel: optional string containing the channel code

    Network: string containing the network code

    Location: optional string containing the location code
    """

    station: str
    channel: Optional[str] = None
    network: str
    location: Optional[str] = None


class Channel(FormatBaseModel):
    """A conversion class used to create, parse, and validate geojson Channel data as part of
    detection data.

    type: string containing the type of this geojson

    geometry: PointGeometry object containing the geojson geometry for this feature

    properties: ChannelProperties object containing the channel properties
    """

    type: str = "Feature"

    geometry: PointGeometry
    properties: ChannelProperties



from typing import Optional

from pydantic import Field

from anssformats.formatbasemodel import FormatBaseModel


class Association(FormatBaseModel):
    """A conversion class used to create, parse, and validate association data as part
    of detection data.

    Attributes
    ----------

    phase: optional string containing the associated phase name

    distance: optional float containing the associated distance in degrees between the
        detection's and the data's locations if associated

    azimuth: optional float containing the associated azimuth in degrees between the
        detection's and the data's locations if associated

    residual: optional float containing the associated residual in seconds of the data
        if associated

    sigma: optional float containing the number of standard deviations of the data from
        the calculated value if associated
    """

    phase: Optional[str] = Field(None, pattern=r"^[A-Za-z]+$")

    distance: Optional[float] = Field(None, ge=0.0)
    azimuth: Optional[float] = Field(None, ge=0.0)
    residual: Optional[float] = None
    sigma: Optional[float] = None



from anssformats.pick import Pick
from anssformats.source import Source
from anssformats.amplitude import Amplitude
from anssformats.analytics import Analytics, Prediction
from anssformats.association import Association
from anssformats.detection import Detection
from anssformats.hypocenter import Hypocenter
from anssformats.magnitude import Magnitude
from anssformats.channel import Channel
from anssformats.filter import Filter
from anssformats.eventType import EventType

__all__ = [
    "Pick",
    "Source",
    "Amplitude",
    "Analytics",
    "Prediction",
    "Association",
    "Detection",
    "Hypocenter",
    "Magnitude",
    "Channel",
    "Filter",
    "EventType",
]



from anssformats.formatbasemodel import FormatBaseModel


class Source(FormatBaseModel):
    """A conversion class used to create, parse, and validate source data as part of
    detection data.

    Attributes
    ----------

    agencyID: string containing the originating agency FDSN ID

    author: string containing the source author
    """

    agencyID: str
    author: str



from datetime import datetime
from typing import Any

from pydantic import BaseModel, GetCoreSchemaHandler, field_validator
from pydantic_core import CoreSchema, core_schema


def convert_datetime_to_iso8601_with_z_suffix(dt: datetime) -> str:
    """Convert provided datetime to an ISO 8601 UTC time string

    Parameters
    ----------
    dt: Datetime containing the date time to convert

    Returns
    -------
    A str containing the ISO 8601 UTC formatted time string
    """
    return dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"


class FormatBaseModel(BaseModel):
    """A Pydantic BaseModel used for any required formatting of keys and values"""

    def model_dump(self):
        """Override the default model_dump method to always exclude None values"""
        return super().model_dump(exclude_none=True)

    def model_dump_json(self):
        """Override the default model_dump_json method to always exclude None values"""
        return super().model_dump_json(exclude_none=True)


class CustomDT(datetime):
    """A convenience class used to strip all datetime objects of timezone information,
    required to bypass Pydantic's automatic inclusion of timezone when parsing JSON
    strings.
    """

    @field_validator("*", mode="before")
    @classmethod
    def validate_no_tz(cls, v: Any, info: core_schema.ValidationInfo) -> Any:
        if isinstance(v, str):
            return datetime.strptime(v, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=None)
        return v.replace(tzinfo=None)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.with_info_before_validator_function(
            cls.validate_no_tz,
            handler(datetime),
            serialization=core_schema.plain_serializer_function_ser_schema(
                convert_datetime_to_iso8601_with_z_suffix,
                when_used="json-unless-none",
            ),
        )



from typing import Optional

from pydantic import Field

from anssformats.formatbasemodel import FormatBaseModel


class Amplitude(FormatBaseModel):
    """A conversion class used to create, parse, and validate amplitude data as part of
    detection data.

    Attributes
    ----------

    amplitude: optional float containing the amplitude

    period: optional float containing the period

    snr: optional float containing the signal to noise ratio, capped at 1E9
    """

    amplitude: Optional[float] = None
    period: Optional[float] = Field(None, ge=0.0)
    snr: Optional[float] = Field(None, ge=0.0, le=1e9)



from typing import List, Literal, Optional

from pydantic import Field


from anssformats.formatbasemodel import CustomDT, FormatBaseModel
from anssformats.eventType import EventType
from anssformats.hypocenter import Hypocenter, HypocenterProperties
from anssformats.source import Source
from anssformats.pick import Pick
from anssformats.magnitude import Magnitude
from anssformats.analytics import Analytics


class Detection(FormatBaseModel):
    """A conversion class used to create, parse, and validate detection data.

    type: string identifying this message as a detection

    id: string containing a unique identifier for this detection

    source: Source object containing the source of the detection

    hypocenter: Hypocenter object containing the hypocenter of the detection

    detectionType: optional string containing the origin type of this detection; valid
        values are "New", "Update", "Final", and "Retract"

    detectionTime: optional datetime containing the time this detection was made

    eventType: optional EventType object containing the event type of the detection

    minimumDistance: optional float containing the distance to the closest station

    rms: optional float containing the detection RMS

    maximumGap: optional float containing the detection gap

    detector: optional string containing the detection grid, algorithm, or other
        information

    pickData: optional list of either Pick objects used to generate
        this detection

    analyticsInfo: optional analytics object containing model output attached to this detection
    """

    type: Literal["Detection"]
    id: str
    source: Source

    hypocenter: Hypocenter

    detectionType: Optional[Literal["New", "Update", "Final", "Retract"]] = None
    detectionTime: Optional[CustomDT] = None

    eventType: Optional[EventType] = None

    minimumDistance: Optional[float] = Field(None, ge=0.0)
    rms: Optional[float] = None
    maximumGap: Optional[float] = Field(None, ge=0.0, le=360.0)

    detector: Optional[str] = None

    pickData: Optional[List[Pick]] = None

    magnitudeData: Optional[List[Magnitude]] = None

    analyticsInfo: Optional[Analytics] = None
